#! /usr/bin/env python
#
"""
Unit tests for `_Machine`:class: and derived classes
(from module `gc3libs.backends.shellcmd`).
"""
# Copyright (C) 2011-2015, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'


import os

from mock import Mock
import pytest

from gc3libs.backends.shellcmd import _LinuxMachine


@pytest.fixture(autouse=True)
def transport():
    transport = Mock()
    yield transport


def test_list_process_tree_empty(transport):
    mach = _LinuxMachine(transport)

    transport.execute_command.return_value = (
        # exit code
        0,
        # stdout
        '',
        # stderr
        '',
    )

    pids = mach.list_process_tree()
    assert pids == []


def test_list_process_tree_full(transport):
    r"""
    Check that the output relative to this process tree is correct
    (process tree as displayed by `ps f`)::

         PID  PPID CMD
        2361  1466 /sbin/upstart --user
        2431  2361  \_ upstart-udev-bridge --daemon --user
        2663  2361  \_ gpg-agent --homedir /home/rmurri/.gnupg --use-standard-socket --daemon
        2679  2361  \_ /usr/lib/at-spi2-core/at-spi-bus-launcher
        2684  2679  |   \_ /usr/bin/dbus-daemon --config-file=/etc/at-spi2/accessibility.conf --nofork --print-address 3
        2725  2361  \_ /usr/bin/lxsession -s Lubuntu -e LXDE
        2736  2725  |   \_ lxpanel --profile Lubuntu
    """
    mach = _LinuxMachine(transport)

    transport.execute_command.return_value = (
        # exit code
        0,
        # stdout
        '''\
2361  1466
2431  2361
2663  2361
2679  2361
2684  2679
2725  2361
2736  2725
        ''',
        # stderr
        '',
    )

    pids = mach.list_process_tree("2361")
    assert pids == [
              # PID   PPID CMD
        "2361", # 2361  1466 /sbin/upstart --user
        "2431", # 2431  2361  \_ upstart-udev-bridge --daemon --user
        "2663", # 2663  2361  \_ gpg-agent --homedir /home/rmurri/.gnupg --use-standard-socket --daemon
        "2679", # 2679  2361  \_ /usr/lib/at-spi2-core/at-spi-bus-launcher
        "2725", # 2684  2679  |   \_ /usr/bin/dbus-daemon --config-file=/etc/at-spi2/accessibility.conf --nofork --print-address 3
        "2684", # 2725  2361  \_ /usr/bin/lxsession -s Lubuntu -e LXDE
        "2736", # 2736  2725  |   \_ lxpanel --profile Lubuntu
    ]


if __name__ == "__main__":
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
#   grdock.py -- Front-end script for running the docking program rDock
#   over a list of ligand files.
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``grdock.py --help`` for program usage
instructions.

Input parameters consists of:

...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2015-02-17:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import grdock
    grdock.GrdockScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_CORES = 1
DEFAULT_MEMORY = Memory(1500,MB)
DEFAULT_WALLTIME = Duration(300,hours)

## custom application class
class GrdockApplication(Application):
    """
    Custom class to wrap the execution of rdock.
    The wrapper script that will be executed on the remote end is
    organised in two steps:
    step 1: cavity creattion (will use rbcavity)
    step 2: docking (will use rbdock)

    Application will take the input ligand file and a ligand index
    ligand index is used to create an output file that maintains the
    same ligand index as a suffix:
    Es: input ligand: Docking1.sd -> output: Docked1.sd
    """
    application_name = 'grbdock'

    def __init__(self, docking_file, docking_index, **extra_args):

        self.output_dir = extra_args['output_dir']
        self.docking_index = docking_index

        inputs = dict()
        outputs = dict()

        grdock_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/grdock_wrapper.sh")

        inputs[grdock_wrapper_sh] = os.path.basename(grdock_wrapper_sh)

        inputs[docking_file] = os.path.basename(docking_file)

        if extra_args['data_folder']:
            for element in os.listdir(extra_args['data_folder']):
                inputs[os.path.abspath(os.path.join(extra_args['data_folder'],
                                                   element))] = os.path.basename(element)

        arguments = "./%s -n %s -o Docked%s %s results" % (inputs[grdock_wrapper_sh],
                                                           extra_args['rbdock_iterations'],
                                                           self.docking_index,
                                                           os.path.basename(docking_file))

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = ["results/"],
            stdout = 'grdock.log',
            join=True,
            **extra_args)


class GrdockScript(SessionBasedScript):
    """
    The script takes as input either a comma separated list of ligand files
    (with `.sd` extension) or a comma separated list of folders were to find the
    input ligand files.
    For each input ligand file, `grdocking` creates an instance of
    GrdockApplication.

    The ``grdock`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``grdock`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GrdockApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GrdockApplication,
            )

    def setup_options(self):
        self.add_param("-i", "--iterations", metavar="[NUM]",
                       dest="rbdock_iterations", default="20",
                       help="Number of iterations for rbdock. "
                       "Default: 20")

        self.add_param("-d", "--data", metavar="[STRING]",
                       dest="data_folder", default=None,
                       help="Path to data folder (e.g. where "
                       "crebbp-without-water-Tripos.mol2, "
                       "Pseudo-Ligand-in-pose.sd, Water-in-3P1C.pdb "
                       "and/or Docking.sd, could be retrieved")

    def parse_args(self):
        """
        Check that each element in args is at least a alid folder
        """

        self.input_dockings = []

        for argument in self.params.args:
            if os.path.isfile(argument) and argument.endswith(".sd"):
                self.input_dockings.append(os.path.abspath(argument))
            if os.path.isdir(argument):
                # walk through it and recursively search for valid .sd files
                for root,dirs,files in os.walk(argument):
                    for ff in files:
                        if ff.endswith(".sd"):
                            self.input_dockings.append(os.path.join(root,ff))

        if self.params.data_folder:
            assert os.path.isdir(self.params.data_folder)


    def new_tasks(self, extra):
        """
        For each input folder, create an instance of GrdockApplication
        """
        tasks = []

        for docking_file in self.input_dockings:

            # extract root folder name to be used as jobname
            jobname = os.path.basename(docking_file)
            res = re.findall('\d+.', docking_file, flags=re.IGNORECASE)
            if res:
                docking_index = res[-1]
            else:
                gc3libs.log.warning("Failed to extract index number from input "
                                    "dockig file %s." % docking_file)
                docking_index = "none"

            extra_args = extra.copy()

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        'run_%s' % jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        'run_%s' % jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        'run_%s' % jobname)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        'run_%s' % jobname)

            extra_args['rbdock_iterations'] = self.params.rbdock_iterations
            extra_args['data_folder'] = self.params.data_folder

            tasks.append(GrdockApplication(
                docking_file,
                docking_index,
                **extra_args))

        return tasks



#! /usr/bin/env python
#
#   gabm.py -- Front-end script for running sheepriver simulation
#   written in java.
#
#   Copyright (C) 2016  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gabm.py --help`` for program usage
instructions.

Input argument consists of:
- hunting range: integer range (e.g. 1:430)
...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-03-29:
  * Initial version
  2016-05-06:
  * TODO: allow 0 as hunting value
  * TODO: check consistency of seed file
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gabm
    gabm.GabmScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_REMOTE_BIN="ABM"

## custom application class
class GabmApplication(Application):
    """
    """
    application_name = 'gabm'

    def __init__(self, hunting, **extra_args):
        """
        Remote execution: ABM [value]
        """

        inputs = dict()
        outputs = dict()
        executables = list()

        if 'binary' in extra_args:
            remote_bin = os.path.basename(extra_args['binary'])
            arguments = "./%s " % remote_bin
            inputs[extra_args['binary']] = remote_bin
            executables.append(remote_bin)
        else:
            arguments = DEFAULT_REMOTE_BIN

        arguments += " %d " % hunting

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'gabm.log',
            join=True,
            executables = executables,
            **extra_args)

class GabmScript(SessionBasedScript):
    """

    The ``gabm`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gabm`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GabmApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GabmApplication,
            )

    def setup_args(self):

        self.add_param('range', type=str,
                       help="hunting pressures range. "
                       "Format: [int],[int]|[int]:[int]. E.g 1:432|3|6,9")

    def setup_options(self):
        self.add_param("-B", "--binary", metavar="STRING", type=str,
                       dest="binary",
                       default=None,
                       help="Location of statically linked binary to run")

    def parse_args(self):
        try:
            if self.params.binary:
                assert os.path.isfile(self.params.binary), \
                    "Input binary file %s not found" % self.params.sources

            # Validate month range
            try:
                # Check whether only single value has been passed
                try:
                    assert isinstance(int(self.params.range),int)
                    self.input_range = [int(self.params.range)]
                except ValueError as ex:
                    # Identify the separator
                    if len(self.params.range.split(":")) == 2:
                        # Use ':' as separator
                        try:
                            start,end = [ int(mrange) for mrange in \
                                          self.params.range.split(":") \
                                          if isinstance(int(mrange),int) ]
                            if end <= start:
                                raise ValueError("No valid input range. "
                                                 "Format: [int],[int]|[int]:[int]. E.g 1:432|3")
                            self.input_range = range(start,end+1)
                        except (TypeError, ValueError) as ex:
                            gc3libs.log.critical(ex.message)
                            raise ValueError("No valid input range. "
                                             "Format: [int],[int]|[int]:[int]. E.g 1:432|3")
                    elif len(self.params.range.split(",")) > 0:
                        # Use ',' as separator
                        self.input_range = [ int(mrange) for mrange in \
                                             self.params.range.split(",") \
                                             if isinstance(int(self.params.range),int) ]
                    else:
                        # Anything else should fail
                        raise ValueError("No valid input range. "
                                         "Format: [int],[int]|[int]:[int]. E.g 1:432|3")

            except ValueError as ex:
                gc3libs.log.debug(ex.message)
                raise AttributeError("No valid input range. "
                                     "Format: [int],[int]|[int]:[int]. E.g 1:432|3")

        except AssertionError as ex:
            raise OSError(ex.message)

    def new_tasks(self, extra):
        """
        For each input folder, create an instance of GabmApplication
        """
        tasks = []

        for hunting in self.input_range:

            # filename example: 0103645_anat.nii.gz
            # extract root folder name to be used as jobname
            extra_args = extra.copy()
            extra_args['jobname'] = 'abm-%s' % str(hunting)

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        'run_%s' % hunting)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        'run_%s' % hunting)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        'run_%s' % hunting)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        'run_%s' % hunting)

            if self.params.binary:
                extra_args['binary'] = os.path.abspath(self.params.binary)

            tasks.append(GabmApplication(
                hunting,
                **extra_args))

        return tasks



# __init__.py
# -*- coding: utf-8 -*-

"""
Sub-package of utility function to aid in writing GC3Pie unit tests.
"""

#  Copyright (C) 2016,  University of Zurich. All rights reserved.
#
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or (at your
#  option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA



#!/bin/env python

"""
# gspg_wrapper.py -- Wrapper script to execute CTX analysis
#
# Authors: Sergio Maffioletti <sergio.maffioletti@uzh.ch>
#
# Copyright (C) 2015-2016  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#-----------------------------------------------------------------------
"""

from __future__ import absolute_import, print_function, unicode_literals
from builtins import next
from builtins import range
import os
import sys
import subprocess
import argparse
import time
import csv

DEFAULT_BINARY="ctx-linkdyn-ordprm-sirs.p4"
RESULT_FILE="results.csv"

def runctx(args):

    results = dict()

    parser = argparse.ArgumentParser(description='Run CTX simulation.')
    parser.add_argument('inputcsv', metavar='I',
                        help='location of input .csv file.')
    parser.add_argument('--ctx', dest='ctx', action='store', default=None,
                        help='Alterantive path to ctx binay (default: None)')

    arguments = parser.parse_args(args)

    command = ""

    if arguments.ctx:
        command = "%s " % arguments.ctx
    else:
        command =  DEFAULT_BINARY
    command += " -i ./input.dat"

    # use csv package only
    with open(arguments.inputcsv,'rb') as rd:
        reader = csv.reader(rd)
        columns = next(reader)
        # open destination file
        for line in reader:
            index_of_dat = line[-1]
            with open("./input.dat",'wb') as fd:
                for index in range(0,len(line)-1):
                    fd.write("%s\t%s\n" % (columns[index],line[index]))

            _process = subprocess.Popen(command,stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        close_fds=True, shell=True)

            stime = time.time()
            (out,err) = _process.communicate()
            ftime = time.time()
            print("Index %s processed in %d" % (index_of_dat,(ftime-stime)))
            exitcode = _process.returncode

            if exitcode == 0:
                results[index_of_dat] = out.strip().split('\t')
            else:
                print("ERROR %d. message: %s" % (exitcode,err))

    # collect all results into a single .csv file
    print("Aggregating results")
    with open(RESULT_FILE,'wb') as rd:
        for idx,line in list(results.items()):
            rd.write(idx + "," + ",".join(x for x in line) + "\n")
    print("Done")

if __name__ == '__main__':
    if len(sys.argv[1:]) < 1:
           sys.exit(1)
    sys.exit(runctx(sys.argv[1:]))



#! /usr/bin/env python
#
"""
Python implementation of rosenbrock.cpp for tests.
"""
# Copyright (C) 2011, 2012, 2013,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
__docformat__ = 'reStructuredText'

import os
import sys
import logging
import tempfile
import shutil
import logging


if __name__ == "__main__":
    print("Compute Rosenbrock function")

    # log = open('/home/benjamin/.gc3/debug.log', 'a')
    # log.write("Running in directory: '%s'\n" % os.getcwd())
    # log.write("Contents of working directory: %s\n" % os.listdir(os.getcwd()))
    # log.close()

    para_file = open("parameters.in")
    x = float(para_file.readline().split()[1])
    y = float(para_file.readline().split()[1])

    fun = 100.0 * (y - x ** 2) ** 2 + (1 - x) ** 2

    out_file = open("rosenbrock.out", 'w')
    out_file.write(str(fun) + '\n')

    print("done")



#! /usr/bin/env python
#
#   gc_gps.py -- Front-end script for submitting multiple `GC-GPS` R-baseed jobs.
#
#   Copyright (C) 2011, 2012, 2019  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `R` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gc_gps.py --help`` for program usage instructions.

Input parameters consists of:
:param str command_file: Path to the file containing all the commands to be executed
   example:
          R CMD BATCH --no-save --no-restore '--args pos=27 realizations=700 snr=1
          mast.h=0.5 sd.mast.o=0' ./src/processit.R ./out/screen.out

:param str src_dir: path to folder containing R scripts to be transferred to
'src' folder on the remote execution node

"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2013-01-24:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@gc3.uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gc_gps
    gc_gps.GcgpsScript().run()

import os
import sys
import time
import tempfile

import shutil

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


## custom application class
class GcgpsApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gc_gps'

    def __init__(self, command, src_dir, result_dir, input_dir, **extra_args):


        # setup output references
        self.result_dir = result_dir
        self.output_dir = extra_args['output_dir']

        # extract from command relevant parameters to build output filename
        command_args = _parse_command(command)
        if not command_args:
            outputs = gc3libs.ANY_OUTPUT
        else:
            self.output_file_name = "pos-%s_rates_MC%s_snr%s_h%s_sd_o%s.Rdata" \
                % (command_args['pos'],
                   command_args['realizations'],
                   command_args['snr'], command_args['mast.h'],
                   command_args['sd.mast.o'])

            self.local_output_file = os.path.join(self.output_dir, self.output_file_name)
            self.local_result_output_file = os.path.join(self.result_dir, self.output_file_name)
            outputs = ['./out/screen.out', ('./out/pos.output', self.output_file_name)]

        # setup input references
        inputs = [ (os.path.join(src_dir,v),os.path.join("./src",v))
                   for v in os.listdir(src_dir)
                   if not os.path.isdir(os.path.join(src_dir,v)) ]

        if input_dir:
            # take everything from input_dir/
            # XXX: N.B. NOT recursively
            for elem in os.listdir(input_dir):
                if os.path.isfile(os.path.join(input_dir,elem)):
                    inputs.append((os.path.join(input_dir, elem),
                                   os.path.join('./in',elem)))

        # prepare execution script from command
        execution_script = """
#!/bin/sh

# Checking 'in' folder
# Checks first if './in' has been created
# otherwise try default in '$HOME/in'

echo -n "Checking Input folder... "
if [ -d ./in ]; then
   echo "[${PWD}/in]"
elif [ -d ${HOME}/in ]; then
   ln -s ${HOME}/in in
   echo [${HOME}/in]
else
   echo [FAILED: no folder found in './in' nor in '${HOME}/in']
   exit 1
fi

mkdir out

# Install docker
echo "Installing dockers... "
if ! command -v docker; then
  curl https://get.docker.com/ | sudo bash
fi
sudo usermod -aG docker $USER

# execute command
echo "Runnning: sudo sudo -u $USER -g docker docker run -i --rm -v "$PWD":/home/docker -w /home/docker -u docker r-base %s"
sudo sudo -u $USER -g docker docker run -i --rm -v "$PWD":/home/docker -w /home/docker -u docker r-base %s
RET=$?

echo Program terminated with exit code $RET

mv out/pos* out/pos.output
exit $RET
        """ % (command, command)

        try:
            # create script file
            (handle, self.tmp_filename) = tempfile.mkstemp(prefix='gc3pie-gc_gps', suffix=extra_args['jobname'])

            # XXX: use NamedTemporaryFile instead with 'delete' = False

            fd = open(self.tmp_filename,'w')
            fd.write(execution_script)
            fd.close()
            os.chmod(fd.name,0o777)
        except Exception, ex:
            gc3libs.log.debug("Error creating execution script" +
                              "Error type: %s." % type(ex) +
                              "Message: %s"  %ex.message)
            raise

        inputs.append((fd.name,'gc_gps.sh'))


        Application.__init__(
            self,
            arguments = ['./gc_gps.sh'],
            inputs = inputs,
            outputs = outputs,
            stdout = 'gc_gps.log',
            join=True,
            **extra_args)

    def fetch_output_error(self, ex):
        # Ignore errors if `pos.output` file has not been created by
        # the application.
        if isinstance(ex, gc3libs.exceptions.CopyError):
            if os.path.basename(ex.source) == 'pos.output':
                return None
        return ex

    def terminated(self):
        """
        Extract output file from 'out'
        """
        # Cleanup tmp file
        try:
            os.remove(self.tmp_filename)
        except Exception, ex:
            gc3libs.log.error("Failed removing temporary file %s. " % self.tmp_filename +
                              "Error type %s. Message %s" % (type(ex), str(ex)))

        if not self.local_output_file:
            # outputs = gc3libs.ANY_OUTPUT
            for path in os.path.listdir(self.output_dir):
                if os.path.isfile(path) and path.startswith('pos'):
                    # We assume this is the output file to retrieve
                    self.local_output_file = path
                    self.local_result_output_file = os.path.join(self.result_dir,path)

        # copy output file `pos*` in result_dir
        if not os.path.isfile(self.local_output_file):
            gc3libs.log.error("Output file %s not found"
                              % self.local_output_file)
            self.execution.returncode = (0, 100)
        else:
            try:
                shutil.move(self.local_output_file,
                            self.local_result_output_file)
            except Exception, ex:
                gc3libs.log.error("Failed while transferring output file " +
                                  "%s " % self.local_output_file +
                                  "to result folder %s. " % self.result_dir +
                                  "Error type %s. Message %s. "
                                  % (type(ex),str(ex)))

                self.execution.returncode = (0, 100)

def _parse_command(command):
        """
        Parse a command like the following:
        "R CMD BATCH --no-save --no-restore '--args pos=5 realizations=700
         snr=1 mast.h=0.5 sd.mast.o=0' ./src/processit.R ./out/screen.out"
        and should return a dictionary like the following:
        {'pos': '5', 'realizations': '700', 'snr': '1',
         'mast.h': '0.5', 'sd.mast.o': '0' }
        """
        try:
            # filter out the part before and after the arguments
            args = command.split("'")[1]

            # split and ignore '--args'
            args = args.split()[1:]

            # return a dictionary from "pos=5 realizations=700 snr=1
            # mast.h=0.5 sd.mast.o=0"
            # return { k:v for k,v in [ v.split('=') for v in args ] }
            return dict( (k,v) for k,v in [ v.split('=') for v in args ])
        except Exception, ex:
            gc3libs.log.error("Failed while parsing command: %s. " % command +
                              "Type: %s. Message: %s" % (type(ex),str(ex)))
            return None

class GcgpsTask(RetryableTask):
    def __init__(self, command, src_dir, result_dir, input_dir, **extra_args):
        RetryableTask.__init__(
            self,
            # actual computational job
            GcgpsApplication(
                command,
                src_dir,
                result_dir,
                input_dir,
                **extra_args),
            **extra_args
            )

    def retry(self):
        """
        Task will be retried iif the application crashed
        due to an error within the exeuction environment
        (e.g. VM crash or LRMS kill)
        """
        # XXX: check whether it is possible to distingish
        # between the error conditions and set meaningfull exitcode
        return False


class GcgpsScript(SessionBasedScript):
    """
Scan the specified INPUT directories recursively for simulation
directories and submit a job for each one found; job progress is
monitored and, when a job is done, its output files are retrieved back
into the simulation directory itself.

A simulation directory is defined as a directory containing a
``geotop.inpts`` file.

The ``ggeotop`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``ggeotop`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GcgpsApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GcgpsApplication,
            )

    def setup_options(self):
        self.add_param("-i", "--input", metavar="PATH", #type=executable_file,
                       dest="input_dir", default=None,
                       help="Path to the input data folder.")

    def setup_args(self):

        self.add_param('command_file', type=str,
                       help="Command file full path name")

        self.add_param('R_source_folder', type=str,
                       help="Path to folder containing scripts to be executed.")

    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """

        # check args:
        # XXX: make them position independent
        if not os.path.isdir(self.params.R_source_folder):
            raise gc3libs.exceptions.InvalidUsage(
                "Invalid path to R scripts folder: '%s'. Path not found"
                % self.params.R_source_folder)
        # XXX: shall we check/validate the content ( presence of valid R scripts ) ?

        self.log.info("source dir: %s" % self.params.R_source_folder)

        if not os.path.exists(self.params.command_file):
            raise gc3libs.exceptions.InvalidUsage(
                "gc_gps command file '%s' does not exist;"
                % self.params.command_file)
        gc3libs.utils.check_file_access(self.params.command_file, os.R_OK,
                                gc3libs.exceptions.InvalidUsage)

        if self.params.input_dir and not os.path.isdir(self.params.input_dir):
            raise gc3libs.exceptions.InvalidUsage(
                "Input folder '%s' does not exists"
                % self.params.input_dir)

        self.log.info("Command file: %s" % self.params.command_file)
        self.log.info("R source dir: %s" % self.params.R_source_folder)
        if self.params.input_dir:
            self.log.info("Input data dir: '%s'" % self.params.input_dir)

    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new GcgpsTask
        """

        tasks = []

        try:
            fd = open(self.params.command_file)

            self.result_dir = os.path.dirname(self.params.output)

            for line in fd:
                command = line.strip()

                if not command:
                    # ignore black lines
                    continue

                cmd_args = _parse_command(command)

                # setting jobname
                jobname = "gc_gps-%s%s%s%s%s" % (cmd_args['pos'],
                                                 cmd_args['realizations'],
                                                 cmd_args['snr'],
                                                 cmd_args['mast.h'],
                                                 cmd_args['sd.mast.o'])

                extra_args = extra.copy()
                extra_args['jobname'] = jobname
                # FIXME: ignore SessionBasedScript feature of customizing
                # output folder
                extra_args['output_dir'] = self.params.output

                extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', os.path.join('.computation',jobname))
                extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', os.path.join('.computation',jobname))
                extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', os.path.join('.computation',jobname))
                extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', os.path.join('.computation',jobname))

                self.log.debug("Creating Task for command: %s" % command)

                tasks.append(GcgpsTask(
                        command,
                        self.params.R_source_folder,
                        self.result_dir,
                        self.params.input_dir,
                        **extra_args))

        except IOError, ioe:
            self.log.error("Error while reading command file " +
                           "%s." % self.params.command_file +
                           "Message: %s" % ioe.message)
        except Exception, ex:
            self.log.error("Unexpected error. Error type: %s, Message: %s" % (type(ex),str(ex)))

        finally:
            fd.close()

        return tasks



#! /usr/bin/env python
#
"""
Test interaction with the SLURM batch-queueing system.
"""
# Copyright (C) 2011-2013, 2015, 2016, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__docformat__ = 'reStructuredText'

import datetime
import os
import tempfile

import mock
import pytest

import gc3libs
import gc3libs.core
import gc3libs.config
from gc3libs.exceptions import LRMSError
from gc3libs.quantity import kB, seconds, minutes

from faketransport import FakeTransport


def sbatch_submit_ok(jobid=123):
    return (
        # command exit code
        0,
        # stdout
        ("Submitted batch job %d\n" % jobid),
        # stderr
        '')

def sbatch_submit_failed(jobid=123):
    return (
        # command exit code
        1,
        # stdout
        '',
        # stderr
        'sbatch: error: Batch job submission failed:'
        ' Invalid account or account/partition combination specified')

def sacct_no_accounting(jobid=123):
    return (
        # command exitcode
        1,
        # stdout
        '',
        # stderr
        'SLURM accounting storage is disabled\n')


def squeue_pending(jobid=123):
    # squeue --noheader --format='GC3Pie^%i^%T^%r' -j $jobid
    return (
        # command exitcode
        0,
        # stdout
        ('GC3Pie^%d^PENDING^Resources' % jobid),
        # stderr
        '')


def squeue_running(jobid=123):
    # squeue --noheader --format='GC3Pie^%i^%T^%r' -j $jobid
    return (
        # command exitcode
        0,
        # stdout
        ('GC3Pie^%d^RUNNING^None' % jobid),
        # stderr
        '')


def squeue_recently_completed(jobid=123):
    # squeue --noheader --format='GC3Pie^%i^%T^%r' -j $jobid
    return (
        # command exitcode
        0,
        # stdout
        '',
        # stderr
        '')


def squeue_pending_with_additions(jobid=123):
    # squeue --noheader --format='GC3Pie^%i^%T^%r' -j $jobid
    return (
        # command exitcode
        0,
        # stdout
        ('Site-specific preamble here.\nGC3Pie^%d^PENDING^Resources\n\nAdditional info.' % jobid),
        # stderr
        '')


def squeue_running_with_additions(jobid=123):
    # squeue --noheader --format='GC3Pie^%i^%T^%r' -j $jobid
    return (
        # command exitcode
        0,
        # stdout
        ('Site-specific preamble here.\nGC3Pie^%d^RUNNING^None\n\nAdditional info.' % jobid),
        # stderr
        '')


def squeue_recently_completed_with_additions(jobid=123):
    # squeue --noheader --format='GC3Pie^%i^%T^%r' -j $jobid
    return (
        # command exitcode
        0,
        # stdout
        'Site-specific output here.',
        # stderr
        '')


def squeue_notfound(jobid=123):
    # squeue --noheader --format='GC3Pie^%i^%T^%r' -j $jobid
    return (
        # command exitcode
        1,
        # stdout
        '',
        # stderr
        'slurm_load_jobs error: Invalid job id specified')


def scancel_success(jobid=123):
    return (0, "", "")


def scancel_notfound(jobid=123):
    return (
        # command exitcode
        1,
        # stdout
        '',
        # stderr
        ('scancel: error: Kill job error on job id %d: '
         'Invalid job id specified' % jobid))


def scancel_permission_denied(jobid=123):
    return (
        # command exitcode (yes, it's really 0!)
        0,
        # stdout
        "",
        # stderr
        "scancel: error: Kill job error on job id %d: "
        "Access/permission denied\n" % jobid)


def sacct_notfound(jobid=123):
    # SLURM 2.6.5 and 14.11.8 on Ubuntu 14.04.2
    return (
        # command exitcode (yes, it's really 0!)
        0,
        # stdout
        "",
        # stderr
        "")


def sacct_done_ok(jobid=123):
    # SLURM 2.6.5 on Ubuntu 14.04.2
    return (
        # command exitcode (yes, it's really 0!)
        0,
        # stdout
        """
{jobid}|0:0|COMPLETED|1|00:08:07|05:05.002|2016-02-16T12:16:33|2016-02-16T14:24:46|2016-02-16T14:32:53|||
{jobid}.batch|0:0|COMPLETED|1|00:08:07|05:05.002|2016-02-16T14:24:46|2016-02-16T14:24:46|2016-02-16T14:32:53|1612088K|7889776K|
        """.strip().format(jobid=jobid),
        # stderr
        "")


def sacct_done_bad_timestamps(jobid=123):
    #    $ sudo sacct --noheader --parsable --format jobid,ncpus,cputimeraw,\
    #            elapsed,submit,start,end,exitcode,maxrss,maxvmsize,totalcpu \
    #            -j 19
    return (
        # command exitcode (yes, it's really 0!)
        0,
        # stdout
        """
{jobid}|0:0|COMPLETED|1|00:01:06|00:00:00|2012-09-24T10:48:34|2012-09-24T10:47:28|2012-09-24T10:48:34|||
{jobid}.0|0:0|COMPLETED|1|00:01:05|00:00:00|2012-09-24T10:47:29|2012-09-24T10:47:29|2012-09-24T10:48:34|0|0|
        """.strip().format(jobid=jobid),
        # stderr
        "")


def sacct_done_parallel(jobid=123):
    # sample gotten from J.-G. Piccinali, CSCS
    return (
        # command exitcode
        0,
        # stdout
        """
{jobid}|0:0|COMPLETED|64|00:00:23|00:01.452|2012-09-04T11:18:06|2012-09-04T11:18:24|2012-09-04T11:18:47|||
{jobid}.batch|0:0|COMPLETED|1|00:00:23|00:01.452|2012-09-04T11:18:24|2012-09-04T11:18:24|2012-09-04T11:18:47|7884K|49184K|
        """.strip().format(jobid=jobid),
        # stderr
        "")

def sacct_done_cancelled(jobid=123):
    return (
        # command exitcode
        0,
        # stdout
        """
{jobid}|0:0|CANCELLED by 1000|4|00:00:05|00:00:00|2014-12-11T17:13:39|2014-12-11T17:13:39|2014-12-11T17:13:44|||
{jobid}.batch|0:15|CANCELLED|1|00:00:05|00:00:00|2014-12-11T17:13:39|2014-12-11T17:13:39|2014-12-11T17:13:44|0|0|
        """.strip().format(jobid=jobid),
        # stderr
        "")

def sacct_done_node_fail(jobid=123):
    return (
        # command exitcode
        0,
        # stdout
        '''
{jobid}|1:0|NODE_FAIL|8|01:21:47|00:00:00|2016-09-23T18:44:37|2016-09-23T18:44:37|2016-09-23T20:06:24|||
{jobid}.0|-2:0|CANCELLED by 1000|8|01:21:47|00:00:00|2016-09-23T18:44:37|2016-09-23T18:44:37|2016-09-23T20:06:24|||
        '''.strip().format(jobid=jobid),
        # stderr
        '')

def sacct_done_fail_early(jobid=123):
    return (
        # command exitcode
        0,
        # stdout
        '''
{jobid}|1:0|FAILED|14|00:00:01|00:00:00|2016-10-14T16:14:49|2016-10-14T16:14:49|2016-10-14T16:14:50|||
{jobid}.batch|1:0|FAILED|14|00:00:01|00:00:00|2016-10-14T16:14:49|2016-10-14T16:14:49|2016-10-14T16:14:50|||
        '''.strip().format(jobid=jobid),
        # stderr
        '')

def sacct_done_timeout(jobid=123):
    return (
        # command exitcode
        0,
        # stdout
        """
{jobid}|0:1|TIMEOUT|4|00:01:11|00:00:00|2014-12-11T17:10:23|2014-12-11T17:10:23|2014-12-11T17:11:34|||
{jobid}.batch|0:15|CANCELLED|1|00:01:11|00:00:00|2014-12-11T17:10:23|2014-12-11T17:10:23|2014-12-11T17:11:34|0|0|
        """.strip().format(jobid=jobid),
        # stderr
        "")

def sacct_done_relative_timestamps(jobid=123):
    # sample gotten from J.-G. Piccinali, CSCS using SLURM_TIME_FORMAT=relative
    return (
        # command exitcode
        0,
        # stdout
        """
{jobid}|0:0|COMPLETED|64|00:00:23|00:01.452|4 Sep 11:18|4 Sep 11:18|4 Sep 11:18|||
{jobid}.batch|0:0|COMPLETED|1|00:00:23|00:01.452|4 Sep 11:18|4 Sep 11:18|\
4 Sep 11:18|7884K|49184K|
        """.strip().format(jobid=jobid),
        # stderr
        "")


def sacct_done_fractional_rusage(jobid=123):
    # sample gotten from Denisa Rodila, University of Geneva
    return (
        # command exitcode
        0,
        # stdout
        """
{jobid}|0:0|COMPLETED|16|00:07:29|58:10.420|2013-08-30T23:16:22|2013-08-30T23:16:22|2013-08-30T23:23:51|||
{jobid}.batch|0:0|COMPLETED|1|00:07:29|00:02.713|2013-08-30T23:16:22|2013-08-30T23:16:22|2013-08-30T23:23:51|62088K|4115516K|
{jobid}.0|0:0|COMPLETED|1|00:06:56|05:44.992|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:26|73784K|401040K|
{jobid}.1|0:0|COMPLETED|1|00:07:01|05:44.968|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:31|74360K|401656K|
{jobid}.2|0:0|COMPLETED|1|00:07:13|05:51.685|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:43|74360K|401720K|
{jobid}.3|0:0|COMPLETED|1|00:07:21|06:01.088|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:51|73644K|401656K|
{jobid}.4|0:0|COMPLETED|1|00:07:16|05:52.315|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:46|69092K|401096K|
{jobid}.5|0:0|COMPLETED|1|00:07:01|05:46.964|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:31|74364K|401104K|
{jobid}.6|0:0|COMPLETED|1|00:07:10|05:46.222|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:40|69148K|401108K|
{jobid}.7|0:0|COMPLETED|1|00:07:10|05:49.074|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:40|74364K|401592K|
{jobid}.8|0:0|COMPLETED|1|00:07:06|05:44.432|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:36|74404K|401688K|
{jobid}.9|0:0|COMPLETED|1|00:07:04|05:45.962|2013-08-30T23:16:30|2013-08-30T23:16:30|2013-08-30T23:23:34|72.50M|401652K|
        """.strip().format(jobid=jobid),
        # stderr
        "")


def sacct_almost_done(jobid=123):
    """All steps completed, but main job still reported as RUNNING."""
    # SLURM 2.6.5 running on Ubuntu 14.04
    return (
        # command exitcode
        0,
        # stdout
        """
{jobid}|0:0|RUNNING|16|00:00:08|00:00.168|2016-03-31T09:33:25|2016-03-31T09:33:25|Unknown|||
{jobid}.batch|0:0|COMPLETED|1|00:00:08|00:00.168|2016-03-31T09:33:25|2016-03-31T09:33:25|2016-03-31T09:33:33|22212K|67032K|
        """.strip().format(jobid=jobid),
        # stderr
        "")


def sacct_done_ok2(jobid=123):
    """All steps and job allocation in state COMPLETED."""
    # SLURM 2.6.5 running on Ubuntu 14.04
    return (
        # command exitcode
        0,
        # stdout
        """
{jobid}|0:0|COMPLETED|16|00:00:08|00:00.168|2016-03-31T09:33:25|2016-03-31T09:33:25|2016-03-31T09:33:33|||
{jobid}.batch|0:0|COMPLETED|1|00:00:08|00:00.168|2016-03-31T09:33:25|2016-03-31T09:33:25|2016-03-31T09:33:33|22212K|67032K|
        """.strip().format(jobid=jobid),
        # stderr
        "")


State = gc3libs.Run.State


class FakeApp(gc3libs.Application):

    def __init__(self):
        gc3libs.Application.__init__(
            self,
            arguments=['/bin/hostname'],               # mandatory
            inputs=[],                  # mandatory
            outputs=[],                 # mandatory
            output_dir="./fakedir",    # mandatory
            stdout="stdout.txt",
            stderr="stderr.txt",
            requested_cores=1,)


class TestBackendSlurm(object):

    CONF = """
[resource/example]
type=slurm
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64
enabled=True
accounting_delay = 5

[auth/ssh]
type=ssh
username=NONEXISTENT
"""

    @pytest.fixture(autouse=True)
    def setUp(self):
        (fd, self.tmpfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        f.write(TestBackendSlurm.CONF)
        f.close()

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(self.tmpfile)

        self.core = gc3libs.core.Core(self.cfg)

        self.backend = self.core.get_backend('example')
        self.backend.transport = FakeTransport()
        self.transport = self.backend.transport
        # basic responses
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['scancel'] = scancel_notfound()

        yield

        os.remove(self.tmpfile)

    def test_sbatch_submit_ok(self):
        """Test `squeue` output parsing with a job in PENDING state."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)
        assert app.execution.state == State.SUBMITTED

    def test_sbatch_submit_failed(self):
        """Test `squeue` output parsing with a job in PENDING state."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_failed()
        with pytest.raises(LRMSError):
            self.core.submit(app)
        #assert_equal(app.execution.state, State.NEW)

    def test_parse_squeue_output_pending(self):
        """Test `squeue` output parsing with a job in PENDING state."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_pending()
        self.core.update_job_state(app)
        assert app.execution.state == State.SUBMITTED

    def test_parse_squeue_output_with_additions_pending(self):
        """Test `squeue` output parsing with a job in PENDING state."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_pending_with_additions()
        self.core.update_job_state(app)
        assert app.execution.state == State.SUBMITTED

    def test_parse_squeue_output_pending_then_running(self):
        """Test `squeue` output parsing with a job in PENDING and then RUNNING state."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_pending()
        self.core.update_job_state(app)
        assert app.execution.state == State.SUBMITTED

        self.transport.expected_answer['squeue'] = squeue_running()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

    def test_parse_squeue_output_with_additions_pending_then_running(self):
        """Test `squeue` output parsing with a job in PENDING and then RUNNING state."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_pending_with_additions()
        self.core.update_job_state(app)
        assert app.execution.state == State.SUBMITTED

        self.transport.expected_answer['squeue'] = squeue_running_with_additions()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

    def test_parse_squeue_output_immediately_running(self):
        """Test `squeue` output parsing with a job that turns immediately to RUNNING state."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_running()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

    def test_parse_squeue_output_with_additions_immediately_running(self):
        """Test `squeue` output parsing with a job that turns immediately to RUNNING state."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_running_with_additions()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

    def test_job_termination1(self):
        """Test that job termination status is correctly reaped if `squeue` fails but `sacct` does not."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_running()
        self.core.update_job_state(app)

        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_done_ok()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

        self._check_parse_sacct_done_ok(app.execution)

    def test_job_termination2(self):
        """Test that job termination status is correctly reaped if neither `squeue` nor `sacct` fails."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_running()
        self.core.update_job_state(app)

        self.transport.expected_answer['squeue'] = squeue_recently_completed()
        self.transport.expected_answer['env'] = sacct_done_ok()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

        self._check_parse_sacct_done_ok(app.execution)

    def test_job_termination3(self):
        """Test that no state update is performed if both `squeue` and `sacct` fail."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_running()
        self.core.update_job_state(app)

        # if the second `sacct` fails, we should not do any update
        self.transport.expected_answer['squeue'] = squeue_recently_completed()
        self.transport.expected_answer['env'] = sacct_notfound()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_done_ok()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

        self._check_parse_sacct_done_ok(app.execution)

    def _check_parse_sacct_done_ok(self, job):
        # common job reporting values (see Issue 78)
        assert job.cores == 1
        assert job.exitcode == 0
        assert job.signal == 0
        assert job.duration == 8*minutes + 7*seconds
        assert job.used_cpu_time == 5*minutes + 5.002*seconds
        assert job.max_used_memory == 7889776 * kB
        # SLURM-specific values
        assert job.slurm_max_used_ram == 1612088 * kB
        assert (job.slurm_submission_time ==
                     datetime.datetime(year=2016,
                                       month=2,
                                       day=16,
                                       hour=12,
                                       minute=16,
                                       second=33))
        assert (job.slurm_start_time ==
                     datetime.datetime(year=2016,
                                       month=2,
                                       day=16,
                                       hour=14,
                                       minute=24,
                                       second=46))
        assert (job.slurm_completion_time ==
                     datetime.datetime(year=2016,
                                       month=2,
                                       day=16,
                                       hour=14,
                                       minute=32,
                                       second=53))

    @mock.patch('gc3libs.backends.batch.time')
    def test_accounting_delay1(self, mock_time):
        """
        Test that no state update is performed if both `squeue` and `sacct` fail repeatedly within the accounting delay.
        """
        mock_time.time.return_value = 0

        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_running()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

        # fail repeatedly within the acct delay, no changes to `app.execution`
        for t in 0, 1, 2:
            mock_time.time.return_value = t
            self.transport.expected_answer['squeue'] = squeue_notfound()
            self.transport.expected_answer['env'] = sacct_notfound()
            self.core.update_job_state(app)
            assert app.execution.state == State.RUNNING
            assert hasattr(app.execution, 'stat_failed_at')
            assert app.execution.stat_failed_at == 0

        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_done_ok()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

    @mock.patch('gc3libs.backends.batch.time')
    def test_accounting_delay2(self, mock_time):
        """
        Test that an error is raised if both `squeue` and `sacct` fail repeatedly, exceeding the accounting delay.
        """
        mock_time.time.return_value = 0

        gc3libs.log.setLevel(10)

        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_running()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

        # fail first within the acct delay, no changes to `app.execution`
        mock_time.time.return_value = 0
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_notfound()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING
        assert hasattr(app.execution, 'stat_failed_at')
        assert app.execution.stat_failed_at == 0

        # fail again, outside the accounting delay
        mock_time.time.return_value = 2000
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_notfound()
        self.core.update_job_state(app)
        assert app.execution.state == State.UNKNOWN

    @mock.patch('gc3libs.backends.batch.time')
    def test_accounting_delay3(self, mock_time):
        """
        Test that no state update is performed if `squeue` and `sacct` disagree on the job status, within the "accounting delay" limit.
        """
        mock_time.time.return_value = 0

        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['squeue'] = squeue_running()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

        # fail repeatedly within the acct delay, no changes to `app.execution`
        for t in 1, 2, 3:
            mock_time.time.return_value = t
            self.transport.expected_answer['squeue'] = squeue_notfound()
            self.transport.expected_answer['env'] = sacct_almost_done()
            self.core.update_job_state(app)
            assert app.execution.state == State.RUNNING
            assert hasattr(app.execution, 'stat_failed_at')
            assert app.execution.stat_failed_at == 1

        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_done_ok2()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

    def test_parse_sacct_output_parallel(self):
        """Test `sacct` output with a successful parallel job."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)
        self.transport.expected_answer['squeue'] = squeue_notfound()
        # self.transport.expected_answer['sacct'] = sacct_done_parallel()
        self.transport.expected_answer['env'] = sacct_done_parallel()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

        job = app.execution
        # common job reporting values (see Issue 78)
        assert job.cores == 64
        assert job.exitcode == 0
        assert job.signal == 0
        assert job.duration == 23 * seconds
        assert job.max_used_memory == 49184 * kB
        assert job.used_cpu_time == 1.452 * seconds
        # SLURM-specific values
        assert job.slurm_max_used_ram == 7884 * kB
        assert (job.slurm_submission_time ==
                     datetime.datetime(year=2012,
                                       month=9,
                                       day=4,
                                       hour=11,
                                       minute=18,
                                       second=6))
        assert (job.slurm_start_time ==
                     datetime.datetime(year=2012,
                                       month=9,
                                       day=4,
                                       hour=11,
                                       minute=18,
                                       second=24))
        assert (job.slurm_completion_time ==
                     datetime.datetime(year=2012,
                                       month=9,
                                       day=4,
                                       hour=11,
                                       minute=18,
                                       second=47))

    def test_parse_sacct_output_bad_timestamps(self):
        """Test `sacct` output with out-of-order timestamps."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)
        self.transport.expected_answer['squeue'] = squeue_notfound()
        # self.transport.expected_answer['sacct'] = sacct_done_bad_timestamps()
        self.transport.expected_answer['env'] = sacct_done_bad_timestamps()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

        job = app.execution
        # common job reporting values (see Issue 78)
        assert job.exitcode == 0
        assert job.signal == 0
        assert job.duration == 66 * seconds
        assert job.max_used_memory == 0 * kB
        assert job.used_cpu_time == 0 * seconds
        # SLURM-specific values
        assert job.slurm_max_used_ram == 0 * kB
        assert (job.slurm_submission_time ==
                     datetime.datetime(year=2012,
                                       month=9,
                                       day=24,
                                       hour=10,
                                       minute=47,
                                       second=28))
        assert (job.slurm_start_time ==
                     datetime.datetime(year=2012,
                                       month=9,
                                       day=24,
                                       hour=10,
                                       minute=47,
                                       second=29))
        assert (job.slurm_completion_time ==
                     datetime.datetime(year=2012,
                                       month=9,
                                       day=24,
                                       hour=10,
                                       minute=48,
                                       second=34))

    def test_parse_sacct_output_fractional_rusage(self):
        """Test `sacct` output with fractional resource usage."""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_done_fractional_rusage()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

        job = app.execution
        # common job reporting values (see Issue 78)
        assert job.exitcode == 0
        assert job.signal == 0
        assert job.duration == 7 * minutes + 29 * seconds
        assert job.max_used_memory == 4115516 * kB
        assert job.used_cpu_time == 58 * minutes + 10.420 * seconds
        # SLURM-specific values
        assert job.slurm_max_used_ram == 74404 * kB
        assert (job.slurm_submission_time ==
                     datetime.datetime(year=2013,
                                       month=8,
                                       day=30,
                                       hour=23,
                                       minute=16,
                                       second=22))
        assert (job.slurm_start_time ==
                     datetime.datetime(year=2013,
                                       month=8,
                                       day=30,
                                       hour=23,
                                       minute=16,
                                       second=22))
        assert (job.slurm_completion_time ==
                     datetime.datetime(year=2013,
                                       month=8,
                                       day=30,
                                       hour=23,
                                       minute=23,
                                       second=51))

    def test_parse_sacct_output_timeout(self):
        """Test `sacct` when job reaches its time limit"""
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_done_timeout()
        self.core.update_job_state(app)
        job = app.execution

        assert job.state ==    State.TERMINATING
        assert job.exitcode == os.EX_TEMPFAIL
        assert job.signal ==   int(gc3libs.Run.Signals.RemoteKill)

    def test_parse_sacct_output_node_fail(self):
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_done_node_fail()
        self.core.update_job_state(app)
        job = app.execution

        assert job.state ==    State.TERMINATING
        assert job.exitcode == os.EX_TEMPFAIL
        assert job.signal ==   int(gc3libs.Run.Signals.RemoteError)

    def test_parse_sacct_output_fail_early(self):
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_done_fail_early()
        self.core.update_job_state(app)
        job = app.execution

        assert job.state ==    State.TERMINATING
        assert job.exitcode == 1
        assert job.signal ==   0

    def test_parse_sacct_output_job_cancelled(self):
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)
        self.transport.expected_answer['squeue'] = squeue_notfound()
        self.transport.expected_answer['env'] = sacct_done_cancelled()
        self.core.update_job_state(app)
        job = app.execution

        assert job.state ==    State.TERMINATING
        assert job.exitcode == os.EX_TEMPFAIL
        assert job.signal ==   int(gc3libs.Run.Signals.RemoteKill)

    def test_cancel_job1(self):
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)

        self.transport.expected_answer['scancel'] = scancel_success()
        self.transport.expected_answer['env'] = sacct_done_cancelled()
        self.core.kill(app)
        assert app.execution.state == State.TERMINATED

    def test_cancel_job2(self):
        app = FakeApp()
        self.transport.expected_answer['sbatch'] = sbatch_submit_ok()
        self.core.submit(app)
        assert app.execution.state == State.SUBMITTED

        self.transport.expected_answer['scancel'] = scancel_permission_denied()
        self.core.kill(app)
        assert app.execution.state == State.TERMINATED

    def test_get_command(self):
        assert self.backend.sbatch == ['sbatch']
        assert self.backend._sacct == 'sacct'
        assert self.backend._scancel == 'scancel'
        assert self.backend._squeue == 'squeue'


def test_get_command_alternate():
    (fd, tmpfile) = tempfile.mkstemp()
    f = os.fdopen(fd, 'w+')
    f.write("""
[auth/ssh]
type=ssh
username=NONEXISTENT

[resource/example]
# mandatory stuff
type=slurm
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64

# alternate command paths
sbatch  = /usr/local/bin/sbatch --constraint=gpu
sacct   = /usr/local/sbin/sacct
scancel = /usr/local/bin/scancel # comments are ignored!
squeue  = /usr/local/bin/squeue
""")
    f.close()

    cfg = gc3libs.config.Configuration()
    cfg.merge_file(tmpfile)
    os.remove(tmpfile)
    b = cfg.make_resources()['example']

    assert b.sbatch == ['/usr/local/bin/sbatch', '--constraint=gpu']
    assert b._sacct == '/usr/local/sbin/sacct'
    assert b._scancel == '/usr/local/bin/scancel'
    assert b._squeue == '/usr/local/bin/squeue'


def test_count_jobs():
    squeue_stdout = '''474867^PENDING^first_user^21913^Priority^(Priority)
474870^PENDING^first_user^21913^Priority^(Priority)
475753^PENDING^first_user^21913^Resources^(Resources)
475747^PENDING^first_user^21913^Resources^(Resources)
475751^PENDING^first_user^21913^Priority^(Priority)
474723^RUNNING^third_user^20345^None^nid0[0002,0005,0013,0017,0022-0023,\
0028-0029,0040,0046,0050,0058-0059,0061,0066,0068-0069,0081,0091,0100,0110,\
0122-0123,0132-0133,0186-0187,0193,0212-0213,0230-0231,0234-0235,0240,\
0276-0277,0280-0281,0294-0295,0298-0299,0302,0314,0325,0332-0333,0335,0337,\
0341,0344-0345,0352-0353,0358-0359,0370-0371,0386-0387,0429,0444-0445,\
0450-0451,0466-0467,0492,0508,0528-0529,0544-0545,0551,0562-0563,0578-0579,\
0608,0628,0636-0637,0642-0644,0651,0654-0655,0658,0670-0673,0684-0685,\
0698-0699,0709,0735,0744-0745,0771,0776-0777,0780-0781,0802-0803,0818,\
0822-0823,0828,0832-0833,0840,0885,0894,0906-0907,0909,0946,0948,0972-0973,\
1011,1022,1024-1025,1052,1059,1074,1098,1100-1103,1106,1122,1132-1133,1137,\
1139,1141-1143,1164-1165,1169,1198-1199,1202,1232-1233,1254-1255,1262,\
1274-1275,1304,1315,1324-1325,1334-1335,1356-1359,1394-1395,1420-1421,1424,\
1455,1459,1488,1492-1493,1519]
475738^RUNNING^first_user^21913^None^nid00[136-137]
475744^RUNNING^first_user^21913^None^nid00[182-183]
475438^RUNNING^second_user^21239^None^nid00[686,720-721,751]
475440^RUNNING^second_user^21239^None^nid00[363,875,916-917]
475448^RUNNING^second_user^21239^None^nid0[0306-0307,1206-1207]
475450^RUNNING^second_user^21239^None^nid0[1263-1265,1296]
475452^RUNNING^second_user^21239^None^nid0[0656,0687,1070,1105]
475736^RUNNING^first_user^21913^None^nid0[0041,1512]
475651^RUNNING^second_user^21239^None^nid00[026,217,867,924]
475726^RUNNING^first_user^21913^None^nid00[742-743]
'''
    # first user has 4 running jobs and 5 queued ones
    R, Q, r, q = gc3libs.backends.slurm.count_jobs(squeue_stdout, 'first_user')
    assert R == 11
    assert Q == 5
    assert r == 4
    assert q == 5
    # second user has 6 running jobs and no queued ones
    R, Q, r, q = gc3libs.backends.slurm.count_jobs(squeue_stdout,
                                                   'second_user')
    assert R == 11
    assert Q == 5
    assert r == 6
    assert q == 0
    # third user has only 1 running job
    R, Q, r, q = gc3libs.backends.slurm.count_jobs(squeue_stdout, 'third_user')
    assert R == 11
    assert Q == 5
    assert r == 1
    assert q == 0


if __name__ == "__main__":
    pytest.main(["-v", __file__])



#! /usr/bin/env python

"""
Job control on PBS/Torque clusters (possibly connecting to the
front-end via SSH).
"""

# Copyright (C) 2009-2014, 2016, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
__docformat__ = 'reStructuredText'


import datetime
import re
import time

from gc3libs import log, Run
from gc3libs.backends import LRMS
import gc3libs.exceptions
from gc3libs.quantity import Memory
from gc3libs.quantity import Duration
from gc3libs.utils import (same_docstring_as, sh_quote_safe_cmdline,
                           sh_quote_unsafe_cmdline)

from . import batch


# data for parsing PBS commands output

# regexps for extracting relevant strings

_qsub_jobid_re = re.compile(r'(?P<jobid>\d+).*', re.I)

_qstat_line_re = re.compile(
    r'^(?P<jobid>\d+)[^\d]+\s+'
    r'(?P<jobname>[^\s]+)\s+'
    r'(?P<username>[^\s]+)\s+'
    r'(?P<time_used>[^\s]+)\s+'
    r'(?P<state>[^\s]+)\s+'
    r'(?P<queue>[^\s]+)')

# convert data to GC3Pie internal format


def _to_memory(val):
    """
    Convert a memory quantity as it appears in the PBS logs to GC3Pie
    `Memory` value.

    Examples::

      >>> from gc3libs.quantity import kB, MB, GB
      >>> _to_memory('44kb') == 44*kB
      True
      >>> _to_memory('12mb') == 12*MB
      True
      >>> _to_memory('2gb') == 2*GB
      True
      >>> _to_memory('1024') == Memory(1024, Memory.B)
      True

    """
    # extract the `kb`, `mb`, etc. suffix, if there is one
    unit = val[-2:]
    # XXX: check that PBS uses base-2 units
    if unit == 'kb':
        return int(val[:-2]) * Memory.kB
    elif unit == 'mb':
        return int(val[:-2]) * Memory.MB
    elif unit == 'gb':
        return int(val[:-2]) * Memory.GB
    elif unit == 'tb':
        return int(val[:-2]) * Memory.TB
    else:
        if val[-1] == 'b':
            # XXX bytes
            val = int(val[:-1])
        else:
            # a pure number
            val = int(val)
        return int(val) * Memory.B


def _parse_asctime(val):
    try:
        # XXX: replace with datetime.strptime(...) in Python 2.5+
        return datetime.datetime(
            *(time.strptime(val, '%m/%d/%Y %H:%M:%S')[0:6]))
    except Exception as err:
        gc3libs.log.error(
            "Cannot parse '%s' as a PBS-format time stamp: %s: %s",
            val, err.__class__.__name__, str(err))
        return None


# code

def count_jobs(qstat_output, whoami):
    """
    Parse PBS/Torque's ``qstat`` output (as contained in string `qstat_output`)
    and return a quadruple `(R, Q, r, q)` where:

      * `R` is the total number of running jobs in the PBS/Torque cell
        (from any user);

      * `Q` is the total number of queued jobs in the PBS/Torque cell
        (from any user);

      * `r` is the number of running jobs submitted by user `whoami`;

      * `q` is the number of queued jobs submitted by user `whoami`
    """
    total_running = 0
    total_queued = 0
    own_running = 0
    own_queued = 0
    for line in qstat_output.split('\n'):
        log.info("Output line: %s" % line)
        m = _qstat_line_re.match(line)
        if not m:
            continue
        if m.group('state') in ['R']:
            total_running += 1
            if m.group('username') == whoami:
                own_running += 1
        elif m.group('state') in ['Q']:
            total_queued += 1
            if m.group('username') == whoami:
                own_queued += 1
        log.info("running: %d, queued: %d" % (total_running, total_queued))

    return (total_running, total_queued, own_running, own_queued)


class PbsLrms(batch.BatchSystem):

    """
    Job control on PBS/Torque clusters (possibly by connecting via SSH
    to a submit node).
    """

    _batchsys_name = 'PBS/TORQUE'

    def __init__(self, name,
                 # this are inherited from the base LRMS class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 auth,  # ignored if `transport` is 'local'
                 # these are inherited from `BatchSystem`
                 frontend, transport,
                 # these are specific to this backend
                 queue=None,
                 # (Note that optional arguments to the `BatchSystem` class,
                 # e.g.:
                 #     keyfile=None, accounting_delay=15,
                 # are collected into `extra_args` and should not be explicitly
                 # spelled out in this signature.)
                 **extra_args):

        # init base class
        batch.BatchSystem.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth,
            frontend, transport,
            **extra_args)

        # backend-specific setup
        self.queue = queue
        self.qsub = self._get_command_argv('qsub')

        # PBS/TORQUE commands
        self._qdel = self._get_command('qdel')
        self._qstat = self._get_command('qstat')
        self._tracejob = self._get_command('tracejob')

    def _parse_submit_output(self, output):
        return self.get_jobid_from_submit_output(output, _qsub_jobid_re)

    def _submit_command(self, app):
        qsub_argv, app_argv = app.qsub_pbs(self)
        if self.queue is not None:
            qsub_argv += ['-q', ('%s' % self.queue)]
        return (sh_quote_safe_cmdline(qsub_argv),
                'cd "$PBS_O_WORKDIR"; ' + sh_quote_unsafe_cmdline(app_argv))

    def _stat_command(self, job):
        return "%s %s | grep ^%s" % (
            self._qstat, job.lrms_jobid, job.lrms_jobid)

    def _acct_command(self, job):
        return "%s %s" % (self._tracejob, job.lrms_jobid)

    def _secondary_acct_command(self, job):
        return "%s -x -f %s" % (self._qstat, job.lrms_jobid)

    def _parse_stat_output(self, stdout, stderr):
        # parse `qstat` output
        pbs_status = stdout.split()[4]
        log.debug("translating PBS/Torque's `qstat` code"
                  " '%s' to gc3libs.Run.State", pbs_status)
        if pbs_status in ['Q', 'W']:
            state = Run.State.SUBMITTED
        elif pbs_status in ['R']:
            state = Run.State.RUNNING
        elif pbs_status in ['S', 'H', 'T'] or 'qh' in pbs_status:
            state = Run.State.STOPPED
        elif pbs_status in ['C', 'E', 'F']:
            state = Run.State.TERMINATING
        else:
            state = Run.State.UNKNOWN
        return self._stat_result(state, None)  # no term status info

    _tracejob_queued_re = re.compile(
        r'(?P<submission_time>\d+/\d+/\d+\s+\d+:\d+:\d+)\s+.\s+'
        r'Job Queued at request of .*job name =\s*(?P<job_name>[^,]+),'
        r'\s+queue =\s*(?P<queue>[^,]+)')

    _tracejob_run_re = re.compile(
        r'(?P<running_time>\d+/\d+/\d+\s+\d+:\d+:\d+)\s+.\s+'
        r'Job Run at request of .*')

    _tracejob_last_re = re.compile(
        r'(?P<end_time>\d+/\d+/\d+\s+\d+:\d+:\d+)\s+.'
        r'\s+Exit_status=(?P<exit_status>\d+)\s+'
        r'resources_used.cput=(?P<used_cpu_time>[^ ]+)\s+'
        r'resources_used.mem=(?P<mem>[^ ]+)\s+'
        r'resources_used.vmem=(?P<used_memory>[^ ]+)\s+'
        r'resources_used.walltime=(?P<used_walltime>[^ ]+)')

    _tracejob_keyval_mapping = {
        # regexp group name
        # |               `Task.execution` attribute
        # |               |
        # |               |                        converter function
        # |               |                        |
        # |               |                        |
        #   ... common backend attrs (see Issue 78) ...
        'exit_status':   ('exitcode',              int),
        'used_cpu_time': ('used_cpu_time',         Duration),
        'used_walltime': ('duration',              Duration),
        'used_memory':   ('max_used_memory',       _to_memory),
        #   ... PBS-only attrs ...
        'mem':           ('pbs_max_used_ram',      _to_memory),
        'submission_time': ('pbs_submission_time', _parse_asctime),
        'running_time':  ('pbs_running_time',      _parse_asctime),
        'end_time':      ('pbs_end_time',          _parse_asctime),
        'queue':         ('pbs_queue',             str),
        'job_name':      ('pbs_jobname',           str),
    }

    def _parse_acct_output(self, stdout, stderr):
        """Parse `tracejob` output."""
        acctinfo = {}
        for line in stdout.split('\n'):
            for pattern, carry_on in [
                    # regexp                   exit loop?
                    # =====================    ==========
                    (self._tracejob_queued_re, True),
                    (self._tracejob_run_re,    True),
                    (self._tracejob_last_re,   False),
            ]:
                match = pattern.match(line)
                if match:
                    for key, value in match.groupdict().items():
                        attr, conv = self._tracejob_keyval_mapping[key]
                        acctinfo[attr] = conv(value)
                    if carry_on:
                        continue
                    else:
                        break
        assert 'exitcode' in acctinfo, (
            "Could not extract exit code from `tracejob` output")
        acctinfo['termstatus'] = Run.shellexit_to_returncode(
            acctinfo.pop('exitcode'))
        return acctinfo

    _pbspro_keyval_mapping = {
        # PBS output key
        # |               `Task.execution` attribute
        # |                          |                  converter function
        # |                          |                  |
        # |                          |                  |
        #   ... common backend attrs (see Issue 78) ...
        'Exit_status':             ('exitcode',         int),
        'resources_used.cpupt':    ('used_cpu_time',    Duration),
        'resources_used.cput':     ('used_cpu_time',    Duration),
        'resources_used.vmem':     ('used_memory',      _to_memory),
        'resources_used.walltime': ('used_walltime',    Duration),
        #   ... PBS-only attrs ...
        'etime':                   ('pbs_queued_at',    _parse_asctime),
        'queue':                   ('pbs_queue',        str),
        'resources_used.mem':      ('pbs_max_used_ram', _to_memory),
        'stime':                   ('pbs_started_at',   _parse_asctime),
    }

    def _parse_secondary_acct_output(self, stdout, stderr):
        """Parse `qstat -x -f` output (PBSPro only)."""
        acctinfo = {}
        # FIXME: could be a bit smarter and not use a dumb quadratic
        # complexity algo...
        for line in stdout.split('\n'):
            for key, (attr, conv) in self._pbspro_keyval_mapping:
                if (key + ' = ') in line:
                    value = line.split('=')[1].strip()
                    acctinfo[attr] = conv(value)
        assert 'exitcode' in acctinfo, (
            "Could not extract exit code from `qstat -x -f` output")
        acctinfo['termstatus'] = Run.shellexit_to_returncode(
            acctinfo.pop('exitcode'))
        return acctinfo

    def _cancel_command(self, jobid):
        return ("%s %s" % (self._qdel, jobid))

    @same_docstring_as(LRMS.get_resource_status)
    @LRMS.authenticated
    def get_resource_status(self):
        self.updated = False
        try:
            self.transport.connect()

            _command = ('%s -a' % self._qstat)
            log.debug("Running `%s`...", _command)
            exit_code, qstat_stdout, stderr \
                = self.transport.execute_command(_command)
            if exit_code != 0:
                # cannot continue
                raise gc3libs.exceptions.LRMSError(
                    "PBS backend failed executing '%s':"
                    " exit code: %d; stdout: '%s', stderr: '%s'"
                    % (_command, exit_code, qstat_stdout, stderr))

            log.debug("Computing updated values for total/available slots ...")
            (total_running, self.queued, self.user_run, self.user_queued) \
                = count_jobs(qstat_stdout, self._username)
            self.total_run = total_running
            self.free_slots = -1
            self.used_quota = -1

            log.info("Updated resource '%s' status:"
                     " free slots: %d,"
                     " total running: %d,"
                     " own running jobs: %d,"
                     " own queued jobs: %d,"
                     " total queued jobs: %d",
                     self.name,
                     self.free_slots,
                     self.total_run,
                     self.user_run,
                     self.user_queued,
                     self.queued,
                     )
            return self

        except Exception as ex:
            # self.transport.close()
            log.error("Error querying remote LRMS, see debug log for details.")
            log.debug("Error querying LRMS: %s: %s",
                      ex.__class__.__name__, str(ex), exc_info=True)
            raise


# main: run tests
if "__main__" == __name__:
    import doctest
    doctest.testmod(name="pbs",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python

"""
This module provides a generic BatchSystem class from which all
batch-like backends should inherit.
"""
# Copyright (C) 2009-2019   University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
__docformat__ = 'reStructuredText'


from collections import namedtuple
from getpass import getuser
import os
import posixpath
import shlex
import sys
import tempfile
import time
import uuid

import gc3libs
from gc3libs import log, Run
import gc3libs.defaults
from gc3libs.backends import LRMS
from gc3libs.utils import same_docstring_as, sh_quote_safe
import gc3libs.backends.transport

# Define some commonly used functions

# FIXME: (Riccardo?) thinks this function is completely wrong and only
# exists to support GAMESS' ``qgms``, which does not allow users to
# specify the name of STDOUT/STDERR files.  When we have a standard
# flexible submission mechanism for all applications, we should remove
# it!


def generic_filename_mapping(jobname, jobid, file_name):
    """
    Map STDOUT/STDERR filenames (as recorded in `Application.outputs`)
    to commonly used default STDOUT/STDERR file names (e.g.,
    ``<jobname>.o<jobid>``).
    """
    try:
        return {('%s.out' % jobname): ('%s.o%s' % (jobname, jobid)),
                ('%s.err' % jobname): ('%s.e%s' % (jobname, jobid)),
                # FIXME: the following is definitely GAMESS-specific
                ('%s.cosmo' % jobname): ('%s.o%s.cosmo' % (jobname, jobid)),
                ('%s.dat' % jobname): ('%s.o%s.dat' % (jobname, jobid)),
                ('%s.inp' % jobname): ('%s.o%s.inp' % (jobname, jobid)),
                ('%s.irc' % jobname): ('%s.o%s.irc' % (jobname, jobid))
                }[file_name]
    except KeyError:
        return file_name


def _make_remote_and_local_path_pair(transport, job, remote_relpath,
                                     local_root_dir, local_relpath):
    """
    Return list of (remote_path, local_path) pairs corresponding to
    """
    # see https://github.com/fabric/fabric/issues/306 about why it is
    # correct to use `posixpath.join` for remote paths (instead of
    # `os.path.join`)
    remote_path = posixpath.join(job.ssh_remote_folder,
                                 generic_filename_mapping(job.lrms_jobname,
                                                          job.lrms_jobid,
                                                          remote_relpath))
    local_path = os.path.join(local_root_dir, local_relpath)
    if transport.isdir(remote_path):
        # recurse, accumulating results
        result = []
        for entry in transport.listdir(remote_path):
            result += _make_remote_and_local_path_pair(
                transport, job,
                posixpath.join(remote_relpath, entry),
                local_path, entry)
        return result
    else:
        return [(remote_path, local_path)]


class BatchSystem(LRMS):

    """
    Base class for backends dealing with a batch-queue system (e.g.,
    PBS/TORQUE, Grid Engine, etc.)

    This is an abstract class, that you should subclass in order to
    interface with a given batch queuing system. (Remember to call
    this class' constructor in the derived class ``__init__`` method.)

    """

    _batchsys_name = 'batch queueing system'
    """
    A human-readable identifier for the batch queueing system
    class/type (e.g., PBS, LSF, etc.).
    """

    def __init__(self, name,
                 # this are inherited from the base LRMS class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 auth,  # ignored if `transport` is 'local'
                 # these are specific to the this backend
                 frontend, transport,
                 accounting_delay=15,
                 # SSH-related options; ignored if `transport` is 'local'
                 ssh_config=None,
                 keyfile=None,
                 ignore_ssh_host_keys=False,
                 ssh_timeout=None,
                 large_file_threshold=None,
                 large_file_chunk_size=None,
                 spooldir=gc3libs.defaults.SPOOLDIR,
                 **extra_args):

        # init base class
        LRMS.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth, **extra_args)

        # backend-specific setup
        self.frontend = frontend
        self.spooldir = spooldir
        if transport == 'local':
            self.transport = gc3libs.backends.transport.LocalTransport()
            self._username = getuser()
        elif transport == 'ssh':
            auth = self._auth_fn()
            self._username=auth.username
            self.transport = gc3libs.backends.transport.SshTransport(
                frontend,
                ignore_ssh_host_keys=ignore_ssh_host_keys,
                ssh_config=(ssh_config or auth.ssh_config),
                username=self._username,
                port=auth.port,
                keyfile=(keyfile or auth.keyfile),
                timeout=(ssh_timeout or auth.timeout),
                large_file_threshold=large_file_threshold,
                large_file_chunk_size=large_file_chunk_size,
            )
        else:
            raise gc3libs.exceptions.TransportError(
                "Unknown transport '%s'" % transport)
        self.accounting_delay = accounting_delay


    def get_jobid_from_submit_output(self, output, regexp):
        """Parse the output of the submission command. Regexp is
        provided by the caller. """
        for line in output.split('\n'):
            match = regexp.match(line)
            if match:
                return match.group('jobid')
        raise gc3libs.exceptions.InternalError(
            "Could not extract jobid from submission command output `{0}`"
            .format(output.rstrip()))

    def _get_command_argv(self, name, default=None):
        """
        Return an *argv*-style array for invoking command `name`.

        The command name is looked up in this resource's configuration
        parameters, and, if found, the associated string is split into
        words (according to the normal POSIX shell rules) and this
        list of words is returned::

          | >>> b = BatchSystem(..., bsub='/usr/local/bin/bsub -R lustre')
          | >>> b._get_command('bsub')
          | ['/usr/local/bin/bsub', '-R', 'lustre']

        Otherwise, if no configuration parameter by name `name` is
        found, then second argument `default` is returned as the sole
        member of the *argv*-array::

          | >>> b = BatchSystem(...)
          | >>> b._get_command('foo', 'bar')
          | ['bar']

        If `default` is `None`, then the value of `name` is used
        instead::

          | >>> b._get_command('baz')
          | ['baz']

        """
        if default is None:
            default = name
        # lookup the command name in the resource config parameters;
        # return it unchanged as a default
        cmd = self.get(name, default)
        # return argv-style array
        return shlex.split(cmd, comments=True)

    def _get_command(self, name, default=None):
        """
        Return an command-line (string) for invoking command `name`.

        The command name is looked up in this resource's configuration
        parameters, and, if found, the associated string is returned::

          | >>> b = BatchSystem(..., bsub='/usr/local/bin/bsub -R lustre')
          | >>> b._get_command('bsub')
          | '/usr/local/bin/bsub -R lustre'

        Otherwise, if no configuration parameter by name `name` is
        found, then second argument `default` is returned::
        member of the *argv*-array::

          | >>> b = BatchSystem(...)
          | >>> b._get_command('foo', 'bar')
          | 'bar'

        If `default` is `None`, then the value of `name` is used
        instead::

          | >>> b._get_command('baz')
          | 'baz'

        """
        if default is None:
            default = name
        # lookup the command name in the resource config parameters;
        # return it unchanged as a default
        argv = self._get_command_argv(name, default)
        return ' '.join(argv)

    def _submit_command(self, app):
        """This method returns a string containing the command to
        issue to submit the job."""
        raise NotImplementedError(
            "Abstract method `_submit_command()` called - "
            "this should have been defined in a derived class.")

    def _parse_submit_output(self, stdout):
        """This method will parse the output of the submit command and
        return the jobid of the submitted job."""
        raise NotImplementedError(
            "Abstract method `parse_submit_output()` called - "
            "this should have been defined in a derived class.")

    def _stat_command(self, job):
        """This method returns a string containing the command to
        issue to get status information about a job."""
        raise NotImplementedError(
            "Abstract method `_stat_command()` called - "
            "this should have been defined in a derived class.")

    def _parse_stat_output(self, stdout):
        """
        Parse the output of the "stat" command and return the current
        status of the job and its exit status.

        :return: A `_stat_result`:class: instance.
        """
        raise NotImplementedError(
            "Abstract method `_parse_stat_output()` called - "
            "this should have been defined in a derived class.")

    # the only purpose of this class definition is to add a docstring
    # to the named tuple
    class _stat_result(namedtuple('_stat_result', ['state', 'termstatus'])):
        """
        Result of parsing the "stat" command output.
        It is a pair ``(state, termstatus)``, where:

        * ``state`` is a valid GC3Pie task state
          (i.e., instance of `Run.State`:class:, which see);

        * ``termstatus`` is either ``None`` (to indicate that no
          termination status information is available), or any value
          that can be used to set GC3Pie's `Run.returncode`:attr: (for
          instance, a 16-bit POSIX process termination status).

        The ``termstatus`` part is only meaningful when ``state``
        is ``TERMINATING``.
        """
        pass

    def _acct_command(self, job):
        """
        Return a string containing the command to issue to get accounting
        information about the `job`.

        It is usually called only if the _stat_command() fails.
        """
        raise NotImplementedError(
            "Abstract method `_acct_command()` called - "
            "this should have been defined in a derived class.")

    def _parse_acct_output(self, stdout):
        """
        Parse the output of `_acct_command` and return a dictionary
        containing infos about the job.

        The `BatchSystem`:class: does not make any assumption about
        the keys contained in the dictionary.
        """
        raise NotImplementedError(
            "Abstract method `_parse_acct_output()` called - "
            "this should have been defined in a derived class.")

    def _secondary_acct_command(self, job):
        """
        Like `_acct_command` but called only if it exits with non-0 status.

        This is used to allow for a fallback method in case there are
        multiple ways to get information from a job.

        By default this method does nothing, as most batch systems do
        only have one and one only way of getting accounting
        information.
        """
        return None

    def _parse_secondary_acct_output(self, stdout):
        """
        Parse the output of `_secondary_acct_command` and return a
        dictionary containing infos about the job.

        The `BatchSystem` class does not make any assumption about
        the keys contained in the dictionary.
        """
        raise NotImplementedError(
            "Abstract method `_parse_secondary_acct_output()` called - "
            "this should have been defined in a derived class.")

    def _cancel_command(self, jobid):
        """This method returns a string containing the command to
        issue to delete the job identified by `jobid`
        """
        raise NotImplementedError(
            "Abstract method `_cancel_command()` called -"
            " this should have been defined in a derived class.")

    def _get_prepost_scripts(self, app, scriptnames):
        script_txt = []
        for script in scriptnames:
            if script not in self:
                gc3libs.log.debug(
                    "%s script not defined for resource %s", script, self.name)
                continue
            if os.path.isfile(self[script]):
                gc3libs.log.debug(
                    "Adding %s file `%s` to the submission script"
                    % (script, self[script]))
                script_file = open(self[script])
                script_txt.append("\n# %s file `%s` BEGIN\n"
                                  % (script, self[script]))
                script_txt.append(script_file.read())
                script_txt.append("\n# %s file END\n" % script)
                script_file.close()
            else:
                # The *content* of the variable will be inserted in
                # the script.
                script_txt.append('\n# inline script BEGIN\n')
                script_txt.append(self[script])
                script_txt.append('\n# inline script END\n')
        return "".join(script_txt)

    def get_prologue_script(self, app):
        """
        This method will get the prologue script(s) for the `app`
        application and will return a string which contains the
        contents of the script(s) merged together.
        """
        prologues = ['prologue', app.application_name + '_prologue',
                     'prologue_content',
                     app.application_name + '_prologue_content']
        return self._get_prepost_scripts(app, prologues)

    def get_epilogue_script(self, app):
        """
        This method will get the epilogue script(s) for the `app`
        application and will return a string which contains the
        contents of the script(s) merged together.
        """
        epilogues = ['epilogue', app.application_name + '_epilogue',
                     'epilogue_content',
                     app.application_name + '_epilogue_content']
        return self._get_prepost_scripts(app, epilogues)

    @LRMS.authenticated
    def submit_job(self, app):
        """This method will create a remote directory to store job's
        sandbox, and will copy the sandbox in there.
        """
        job = app.execution

        # Create the remote directory.
        self.transport.connect()
        cmd = (
            "mkdir -p {0};"
            " mktemp -d {0}/batch_job.XXXXXXXXXX"
            .format(self.spooldir))
        exit_code, stdout, stderr = self.transport.execute_command(cmd)
        if exit_code != 0:
            raise gc3libs.exceptions.SpoolDirError(
                "Cannot create temporary job working directory"
                " on resource '%s'; command '%s' exited"
                " with code: %d and stderr: '%s'."
                % (self.name, cmd, exit_code, stderr))
        ssh_remote_folder = stdout.split('\n')[0]

        # Copy the input file(s) to remote directory.
        for local_path, remote_path in list(app.inputs.items()):
            remote_path = os.path.join(ssh_remote_folder, remote_path)
            remote_parent = os.path.dirname(remote_path)
            try:
                if remote_parent not in ['', '.']:
                    log.debug("Making remote directory '%s'",
                              remote_parent)
                    self.transport.makedirs(remote_parent)
                log.debug("Transferring file '%s' to '%s'",
                          local_path.path, remote_path)
                self.transport.put(local_path.path, remote_path)
                # preserve execute permission on input files
                if os.access(local_path.path, os.X_OK):
                    self.transport.chmod(remote_path, 0o755)
            except:
                log.critical(
                    "Copying input file '%s' to remote cluster '%s' failed",
                    local_path.path, self.frontend)
                raise

        if app.arguments[0].startswith('./'):
            gc3libs.log.debug("Making remote path '%s' executable.",
                              app.arguments[0])
            self.transport.chmod(os.path.join(ssh_remote_folder,
                                              app.arguments[0]), 0o755)

        # if STDOUT/STDERR should be saved in a directory, ensure it
        # exists (see Issue 495 for details)
        for dest in (app.stdout, app.stderr):
            if dest:
                destdir = os.path.dirname(dest)
                if destdir:
                    self.transport.makedirs(
                        posixpath.join(ssh_remote_folder, destdir))

        try:
            sub_cmd, aux_script = self._submit_command(app)
            if aux_script != '':
                # create temporary script name
                script_filename = ('./script.%s.sh' % uuid.uuid4())
                # save script to a temporary file and submit that one instead
                local_script_file = tempfile.NamedTemporaryFile(mode='wt')
                local_script_file.write('#!/bin/sh\n')
                # Add preamble file
                prologue = self.get_prologue_script(app)
                if prologue:
                    local_script_file.write(prologue)

                local_script_file.write(aux_script)

                # Add epilogue files
                epilogue = self.get_epilogue_script(app)
                if epilogue:
                    local_script_file.write(epilogue)

                local_script_file.flush()
                # upload script to remote location
                self.transport.put(
                    local_script_file.name,
                    os.path.join(ssh_remote_folder, script_filename))
                # set execution mode on remote script
                self.transport.chmod(
                    os.path.join(ssh_remote_folder, script_filename), 0o755)
                # cleanup
                local_script_file.close()
                if os.path.exists(local_script_file.name):
                    os.unlink(local_script_file.name)
            else:
                # we still need a script name even if there is no
                # script to submit
                script_filename = ''

            # Submit it
            exit_code, stdout, stderr = self.transport.execute_command(
                "/bin/sh -c %s" % sh_quote_safe('cd %s && %s %s' % (
                    ssh_remote_folder, sub_cmd, script_filename)))

            if exit_code != 0:
                raise gc3libs.exceptions.LRMSError(
                    "Failed executing command 'cd %s && %s %s' on resource"
                    " '%s'; exit code: %d, stderr: '%s'."
                    % (ssh_remote_folder, sub_cmd, script_filename,
                       self.name, exit_code, stderr))

            jobid = self._parse_submit_output(stdout)
            log.debug('Job submitted with jobid: %s', jobid)

            job.execution_target = self.frontend

            job.lrms_jobid = jobid
            job.lrms_jobname = jobid
            try:
                if app.jobname:
                    job.lrms_jobname = app.jobname
            except:
                pass

            if 'stdout' in app:
                job.stdout_filename = app.stdout
            else:
                job.stdout_filename = '%s.o%s' % (job.lrms_jobname, jobid)
            if app.join:
                job.stderr_filename = job.stdout_filename
            else:
                if 'stderr' in app:
                    job.stderr_filename = app.stderr
                else:
                    job.stderr_filename = '%s.e%s' % (job.lrms_jobname, jobid)
            job.history.append('Submitted to %s @ %s, got jobid %s'
                               % (self._batchsys_name, self.name, jobid))
            job.history.append("Submission command output:\n"
                               "  === stdout ===\n%s"
                               "  === stderr ===\n%s"
                               "  === end ===\n"
                               % (stdout, stderr), 'pbs', 'qsub')
            job.ssh_remote_folder = ssh_remote_folder

            return job

        except:
            log.critical(
                "Failure submitting job to resource '%s' - "
                "see log file for errors", self.name)
            raise


    def __run_command_and_parse_output(self, cmd, parser, kind='accounting'):
        log.debug("Checking remote job %s info with `%s` ...", kind, cmd)
        exit_code, stdout, stderr = self.transport.execute_command(cmd)
        if exit_code == 0:
            return parser(stdout, stderr)
        else:
            raise gc3libs.exceptions.AuxiliaryCommandError(
                "Failed running %s command `%s`:"
                " exit code: %d, stderr: '%s'"
                % (kind, cmd, exit_code, stderr),
                do_log=True)


    @same_docstring_as(LRMS.update_job_state)
    @LRMS.authenticated
    def update_job_state(self, app):
        job = app.execution
        try:
            job.lrms_jobid
        except AttributeError as ex:
            # `job` has no `lrms_jobid`: object is invalid
            raise gc3libs.exceptions.InvalidArgument(
                "Job object is invalid: {ex}".format(ex=ex))

        self.transport.connect()

        cmd = self._stat_command(job)
        try:
            state, termstatus = self.__run_command_and_parse_output(
                cmd, self._parse_stat_output, 'status')
            if state != Run.State.TERMINATING:
                # no need to go further and parse acct info; also,
                # exit status is not relevant in this case
                job.state = state
                log.debug("Task %s state set to %s", app, state)
                return state
        except gc3libs.exceptions.AuxiliaryCommandError:
            # use the special state value ``None`` to signal that
            # the "status" command failed, we might need this
            # after the "acct" command has run
            state, termstatus = None, None
        assert state is None or state == Run.State.TERMINATING
        log.debug("Job status command gave state `%s`"
                  " and termination status `%s` for task %s",
                  state, termstatus, app)

        # In some batch systems, jobs disappear from qstat
        # output as soon as they are finished. In these cases,
        # we have to check some *accounting* command to check
        # the exit status.
        acctinfo = {}
        for cmd_fn, parse_fn in [
                # this is the regular sacct/qacct/bjobs command
                (self._acct_command, self._parse_acct_output),
                # This is used to distinguish between a standard
                # Torque installation and a PBSPro where `tracejob`
                # does not work but if `job_history_enable=True`,
                # then we can actually access information about
                # finished jobs with `qstat -x -f`.
                (self._secondary_acct_command, self._parse_secondary_acct_output),
        ]:
            cmd = cmd_fn(job)
            # `._secondary_acct_command` returns ``None`` if no
            # "secondary" accouting method is defined -- skip to next
            # iteration, if any
            if cmd is None:
                continue
            try:
                acctinfo = self.__run_command_and_parse_output(
                    cmd, parse_fn, 'accounting')
                # use info from the first acct command that succeeds
                if acctinfo:
                    log.debug("Gathered accounting info %r for task %s",
                              acctinfo, app)
                    break
            except gc3libs.exceptions.AuxiliaryCommandError:
                log.debug("Accounting command `%s` failed.", cmd)
                # try next one
                pass
            except gc3libs.exceptions.UnexpectedJobState as ex:
                log.debug(
                    "Unexpected output from accounting command `%s`: %s.",
                    cmd, ex)
                # try next one
                pass

        # if no termination status is known and the acct
        # command provided one, use it
        if 'termstatus' in acctinfo:
            # if we have a termination status, then the job has terminated
            state = Run.State.TERMINATING
            if termstatus is None:
                termstatus = acctinfo['termstatus']
            else:
                # this should not happen!  but one never knows how new
                # versions of the software may break old habits and
                # parsing rules, so better fail loudly here so we get
                # a bug report and a chance to fix...
                assert termstatus == acctinfo['termstatus'], (
                    "Status and accounting commands disagree"
                    " on job termination status: {1} vs {2}"
                    .format(termstatus, acctinfo['termstatus'])
                )

        if termstatus is None:
            # No *stat command and no *acct command returned correctly.
            try:
                job.stat_failed_at
            except AttributeError:
                # this is the first time `qstat` fails, record a
                # timestamp and retry later
                job.stat_failed_at = time.time()
                return job.state

            if (time.time() - job.stat_failed_at) <= self.accounting_delay:
                # do nothing, let's try later...
                return job.state
            else:
                # accounting info should be there, if it's not
                # then job is definitely lost
                job.state = Run.State.UNKNOWN
                raise gc3libs.exceptions.LRMSError(
                    "Could not retrieve status information for task {app}"
                    .format(app=app))

        # if we got to this point the job is in TERMINATING state
        # and we know at least the termination status
        assert state == Run.State.TERMINATING

        job.state = state
        job.returncode = termstatus
        job.update(acctinfo)

        return state

    @same_docstring_as(LRMS.peek)
    @LRMS.authenticated
    def peek(self, app, remote_filename, local_file, offset=0, size=None):
        job = app.execution
        assert 'ssh_remote_folder' in job, \
            "Missing attribute `ssh_remote_folder` on `Job` instance" \
            " passed to `PbsLrms.peek`."

        if size is None:
            size = sys.maxsize

        _filename_mapping = generic_filename_mapping(
            job.lrms_jobname, job.lrms_jobid, remote_filename)
        _remote_filename = os.path.join(
            job.ssh_remote_folder, _filename_mapping)

        try:
            self.transport.connect()
            remote_handler = self.transport.open(
                _remote_filename, mode='r', bufsize=-1)
            remote_handler.seek(offset)
            data = remote_handler.read(size)
        except Exception as ex:
            log.error("Could not read remote file '%s': %s: %s",
                      _remote_filename, ex.__class__.__name__, str(ex))

        try:
            local_file.write(data)
        except (TypeError, AttributeError):
            output_file = open(local_file, 'w+b')
            output_file.write(data)
            output_file.close()
        log.debug('... Done.')

    @same_docstring_as(LRMS.validate_data)
    def validate_data(self, data_file_list):
        """
        Supported protocols: file
        """
        for url in data_file_list:
            if url.scheme not in ['file']:
                return False
        return True

    @same_docstring_as(LRMS.cancel_job)
    @LRMS.authenticated
    def cancel_job(self, app):
        job = app.execution
        try:
            self.transport.connect()
            cmd = self._cancel_command(job.lrms_jobid)
            exit_code, stdout, stderr = self.transport.execute_command(cmd)
            if exit_code != 0:
                # XXX: It is possible that 'qdel' fails because job
                # has been already completed thus the cancel_job
                # behaviour should be tolerant to these errors.
                log.error(
                    "Failed executing remote command '%s'; exit status %d",
                    cmd, exit_code)
                log.debug("  remote command returned STDOUT '%s'", stdout)
                log.debug("  remote command returned STDERR '%s'", stderr)
                if exit_code == 127:
                    # command was not executed, time to signal an exception
                    raise gc3libs.exceptions.LRMSError(
                        "Cannot execute remote command '%s'"
                        " -- See DEBUG level log for details" % (cmd,))
            return job
        except:
            log.critical('Failure checking status')
            raise

    @same_docstring_as(LRMS.free)
    @LRMS.authenticated
    def free(self, app):

        job = app.execution
        try:
            self.transport.connect()
            self.transport.remove_tree(job.ssh_remote_folder)
        except Exception as err:
            log.warning("Failed removing remote folder '%s': %s: %s",
                        job.ssh_remote_folder, err.__class__, err)
        return

    @same_docstring_as(LRMS.get_results)
    @LRMS.authenticated
    def get_results(self, app, download_dir,
                    overwrite=False, changed_only=True):
        if app.output_base_url is not None:
            raise gc3libs.exceptions.UnrecoverableDataStagingError(
                "Retrieval of output files to non-local destinations"
                " is not supported (yet).")

        job = app.execution
        try:
            self.transport.connect()
            # Make list of files to copy, in the form of (remote_path,
            # local_path) pairs.  This entails walking the
            # `Application.outputs` list to expand wildcards and
            # directory references.
            stageout = list()
            for remote_relpath, local_url in app.outputs.items():
                local_relpath = local_url.path
                if remote_relpath == gc3libs.ANY_OUTPUT:
                    remote_relpath = ''
                    local_relpath = ''
                stageout += _make_remote_and_local_path_pair(
                    self.transport, job, remote_relpath, download_dir,
                    local_relpath)

            # copy back all files, renaming them to adhere to the
            # ArcLRMS convention
            log.debug("Downloading job output into '%s' ...", download_dir)
            for remote_path, local_path in stageout:
                # ignore missing files (this is what ARC does too)
                self.transport.get(remote_path, local_path,
                                   ignore_nonexisting=True,
                                   overwrite=overwrite,
                                   changed_only=changed_only)
            return

        except:
            raise

    @same_docstring_as(LRMS.validate_data)
    @LRMS.authenticated
    def close(self):
        self.transport.close()


if "__main__" == __name__:
    import doctest
    doctest.testmod(name="sge",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2017-2019,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
__docformat__ = 'reStructuredText'

import os
import gc3libs
from gc3libs.cmdline import SessionBasedDaemon


class SimpleDaemon(SessionBasedDaemon):
    """Simple daemon"""
    version = '1.0'

    def new_tasks(self, extra):
        """
        Populate session with initial tasks.
        """
        jobname = 'EchoApp'
        extra['output_dir'] = os.path.join(self.params.working_dir, jobname)
        return [
            gc3libs.Application(
                ['/bin/echo', 'first run'],
                inputs=[],
                outputs=gc3libs.ANY_OUTPUT,
                stdout='stdout.txt',
                stderr='stderr.txt',
                jobname=jobname,
                **extra)
            ]

    def created(self, inbox, subject):
        """
        A new file has been created. Process it.
        """
        path = subject.path
        jobname = ('LSApp.' + os.path.basename(path))
        extra = self.extra.copy()
        extra['output_dir'] = os.path.join(self.params.working_dir, jobname)
        self.add(
            gc3libs.Application(
                ['/bin/echo', path],
                inputs={path:'foo'},
                outputs=gc3libs.ANY_OUTPUT,
                stdout='stdout.txt',
                stderr='stderr.txt',
                jobname=jobname,
                **extra)
        )


## main: run tests

if "__main__" == __name__:
    from simpledaemon import SimpleDaemon
    SimpleDaemon().run()



#! /usr/bin/env python
#
#   gmbsim.py -- GC3Pie front-end for running the
#   "sim_run_dclone_design_test.R" by Mollie Brooks
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for running multiple `sim_run_dclone_design_test` instances.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gmbsim --help`` for program usage instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2014-03-03:
    * Initial release, forked off the ``gmhc_coev`` sources.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '1.0.0'

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == '__main__':
    import gmbsim
    gmbsim.GmbsimScript().run()


# std module imports
import csv
from itertools import product
import os
import re
import sys
import time
from pkg_resources import Requirement, resource_filename

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, existing_file
from collections import defaultdict
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


## utilities

def repr_as_R(val):
    if isinstance(val, (list, tuple)):
        return ('c(' + str.join(',', [repr_as_R(v) for v in val]) + ')')
    elif isinstance(val, (str, unicode)):
        return val
    else:
        return repr(val)

def grep_not_matching(it, ignore=re.compile(r'^\s*(#.*)?$')):
    """
    Return items in `it` that do *not* match the `ignore` regexp.

    First argument `it` should be an iterator over string values.
    Second optional argument is a regexp that will be searched for
    in every value: values that match will be skipped.
    """
    while True:
        item = next(it)
        if not ignore.search(item):
            yield item


## custom application class

class GmbsimApplication(Application):
    """
    Custom class to wrap the execution of the ``sim_run_dclone_design_test``
    code by M. Brooks.
    """

    application_name = 'sim_run_dclone_design_test'

    def __init__(self,
                 scriptfile,   # path to the R script to run
                 datafiles,    # additional files to upload
                 days_of_the_week, # list of up to 7 int
                 sampling_exp, # list of 3 float
                 isolation_exp,# list of 4 int
                 detection_exp,# list of 7 float
                 nb, # see `R2jags::jags` param `n.burnin`
                 ni, # see `R2jags::jags` param `n.iter`
                 nt, # see `R2jags::jags` param `n.thin`
                 **extra_args):
        # use a wrapper script to drive remote run
        wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                       "gc3libs/etc/run_R.sh")
        inputs = { wrapper_sh:os.path.basename(wrapper_sh) }
        # upload R script
        exename = os.path.basename(scriptfile)
        run_name = './' + exename
        inputs[scriptfile] = exename
        # upload additional files
        for path in datafiles:
            inputs[path] = os.path.basename(path)
        # save command-line params
        self.days_of_the_week = days_of_the_week
        self.sampling_exp = sampling_exp
        self.isolation_exp = isolation_exp
        self.detection_exp = detection_exp
        self.nb = nb
        self.ni = ni
        self.nt = nt
        # provide defaults for envelope requests
        extra_args.setdefault('requested_cores',        4)
        extra_args.setdefault('requested_memory',       1*GB)
        extra_args.setdefault('requested_architecture', Run.Arch.X86_64)
        extra_args.setdefault('requested_walltime',     12*hours)
        # chain into `Application` superclass ctor
        Application.__init__(
            self,
            arguments=( #['/bin/echo'] +
                        ['./' + os.path.basename(wrapper_sh), run_name]
                        + [ repr_as_R(val) for val in (
                            days_of_the_week,
                            sampling_exp,
                            isolation_exp,
                            detection_exp,
                            nb,
                            ni,
                            nt,
                            extra_args['requested_cores'], # nc
                        )]),
            inputs = inputs,
            outputs = [
                'dclone_design_test_fits.Rdata',
                self.application_name + '.log',
            ],
            stdout = self.application_name + '.log',
            join=True,
            **extra_args)


## main script class

class GmbsimScript(SessionBasedScript):
    """
Read the specified INPUT ``.csv`` files and submit jobs according
to the content of those files.

The ``gmbsim`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gmbsim`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            input_filename_pattern = '*.csv',
            application = GmbsimApplication,
            # only display stats for the top-level policy objects
            #stats_only_for = GmbsimApplication,
            )


    def setup_options(self):
        self.add_param("-b", "-nb", "--burn-in", metavar="NUM",
                       dest="nb", default=1,
                       help="Execute NUM iterations for JAGS burn-in.")
        self.add_param("-i", "-ni", "--iter", metavar="NUM",
                       dest="ni", default=3,
                       help="Execute NUM JAGS main iterations.")
        self.add_param("-t", "-nt", "--thin", "--thinning", metavar="NUM",
                       dest="nt", default=1,
                       help="Execute NUM iterations for JAGS thinning.")
        # change default for the core/memory/walltime options
        self.actions['memory_per_core'].default = 1*Memory.GB
        self.actions['wctime'].default = '2 hours'
        self.actions['ncores'].default = 4



    def setup_args(self):
        self.add_param('replicates', type=int,
                       help="Number of replicates to run,"
                       " for each combination of experiment parameters.")
        self.add_param('sampling', type=existing_file,
                       help="File containing sampling experiment definition,"
                       " one per line.")
        self.add_param('isolation', type=existing_file,
                       help="File containing isolation experiment definition,"
                       " one per line.")
        self.add_param('detection', type=existing_file,
                       help="File containing detection experiment definition,"
                       " one per line.")
        self.add_param("scriptfile", type=existing_file, default=None,
                       help="Path to the `sim_run_dclone_design_test` script source.")
        self.add_param("datafiles", nargs='*', type=existing_file, default=[],
                       help="Additional data files to upload, e.g., `weird_dates_sdur.R`.")


    def make_directory_path(self, pathspec, jobname):
        # XXX: Work around SessionBasedScript.process_args() that
        # apppends the string ``NAME`` to the directory path.
        # This is really ugly, but the whole `output_dir` thing needs to
        # be re-thought from the beginning...
        if pathspec.endswith('/NAME'):
            return pathspec[:-len('/NAME')]
        else:
            return pathspec


    @staticmethod
    def read_param_file(filename):
        with open(filename, 'r') as stream:
            csvdata = csv.reader(grep_not_matching(stream))
            return [ tuple(item.replace(' ','') for item in row)
                     for row in csvdata ]


    def new_tasks(self, extra):
        dates_and_sampling_exps = self.read_param_file(self.params.sampling)
        isolation_exps = self.read_param_file(self.params.isolation)
        detection_exps = self.read_param_file(self.params.detection)

        for date_and_sampling, isolation, detection in product(dates_and_sampling_exps, isolation_exps, detection_exps):
            dates = date_and_sampling[0]
            sampling = date_and_sampling[1:]
            extra_args = extra.copy()
            # identify job uniquely -- replace forward slashes as this
            # will be part of a directory/file name
            basename = (str
                        .join('_', [GmbsimApplication.application_name]
                              + [ (param +'='+ repr_as_R(locals()[param]))
                                  for param in ('dates', 'sampling',
                                                'isolation', 'detection') ])
                        .replace('/', 'over'))

            # prepare job(s) to submit
            already = len([ task for task in self.session
                            if ((task.dates, task.sampling, task.isolation, task.detection)
                                == (dates, sampling, isolation, detection)) ])
            base_output_dir = extra_args.pop('output_dir', self.params.output)
            for n in range(already, self.params.replicates):
                jobname=('%s#%d' % (basename, n+1))
                yield GmbsimApplication(
                    scriptfile=self.params.scriptfile,
                    datafiles=self.params.datafiles,
                    days_of_the_week=dates,
                    sampling_exp=sampling,
                    isolation_exp=isolation,
                    detection_exp=detection,
                    nb=self.params.nb,
                    ni=self.params.ni,
                    nt=self.params.nt,
                    jobname=jobname,
                    output_dir=os.path.join(base_output_dir, jobname),
                    **extra_args)
